const std = @import("std");

const nds9 = @import("nds9.zig");
const nds7 = @import("nds7.zig");

const PriorityQueue = std.PriorityQueue(Event, void, Event.lessThan);
const Allocator = std.mem.Allocator;

tick: u64 = 0,
queue: PriorityQueue,

pub fn init(allocator: Allocator) !@This() {
    var queue = PriorityQueue.init(allocator, {});
    try queue.add(.{ .tick = std.math.maxInt(u64), .kind = .heat_death });

    return .{ .queue = queue };
}

pub fn push(self: *@This(), kind: Event.Kind, offset: u64) void {
    self.queue.add(.{ .kind = kind, .tick = self.tick + offset }) catch unreachable;
}

pub inline fn peek(self: *const @This()) Event {
    @setRuntimeSafety(false);

    return self.queue.items[0];
}

pub fn deinit(self: @This()) void {
    self.queue.deinit();
}

pub fn now(self: @This()) u64 {
    return self.tick;
}

pub fn next(self: @This()) u64 {
    @setRuntimeSafety(false);

    return self.queue.items[0].tick;
}

pub fn reset(self: *@This()) void {
    self.tick = 0;
}

pub fn handle7(self: *@This(), bus: *nds7.Bus) void {
    _ = bus;
    const event = self.queue.remove();
    // const late = self.tick - event.tick;

    switch (event.kind) {
        .nds7 => {},
        .heat_death => unreachable,
        .nds9 => unreachable, // handled by handle7
    }
}

pub fn handle9(self: *@This(), bus: *nds9.Bus) void {
    const event = self.queue.remove();
    const late = self.tick - event.tick;

    switch (event.kind) {
        .nds9 => |ev| switch (ev) {
            .draw => {
                bus.ppu.drawScanline(bus);
                bus.ppu.onHdrawEnd(self, late);
            },
            .hblank => bus.ppu.onHblankEnd(self, late),
            .vblank => bus.ppu.onHblankEnd(self, late),
        },
        .heat_death => unreachable,
        .nds7 => unreachable, // handled by handle7
    }
}

pub const Event = struct {
    tick: u64,
    kind: Kind,

    pub const Event7 = enum {};

    pub const Event9 = enum {
        draw,
        hblank,
        vblank,
    };

    pub const Kind = union(enum) {
        heat_death: void,
        nds9: Event9,
        nds7: Event7,
    };

    fn lessThan(_: void, left: @This(), right: @This()) std.math.Order {
        return std.math.order(left.tick, right.tick);
    }
};
